//    Copyright (C) 2004 Michael Mahr

//    This file is part of "PS2L_UtilClasses".
//
//    "PS2L_UtilClasses" is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    "PS2L_UtilClasses" is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with "PS2L_UtilClasses"; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "matrixmanager.h"

#include <iostream>

MatrixManager::MatrixManager()
{
  //fill all those matrices with the identity-matrix
  identity(modelviewmatrix[0]);
  memcpy(modelviewtransform, modelviewmatrix, sizeof(float) * 16); //copy over to first transform
  memcpy(perspectivematrix, modelviewmatrix, sizeof(float) * 16); //copy over to perspective matrix
  memcpy(screenfitmatrix, modelviewmatrix, sizeof(float) * 16); //copy over to screenfitmatrix
  memcpy(perspectivescreenfittransform, modelviewmatrix, sizeof(float) * 16); //copy over to perspectivescreenfittransform
  memcpy(clipspacetransform, modelviewmatrix, sizeof(float) * 16); //copy over to clipspacetransform
  memcpy(finaltransform, modelviewmatrix, sizeof(float) * 16); //copy over to final

  modelviewtop = 0;
  
  finalchanged = false;
  clipspacechanged = false;
}

void MatrixManager::modelviewPush(float * matrix)
{
  modelviewtop ++;
  memcpy(modelviewmatrix[modelviewtop], matrix, sizeof(float) * 16);
  matrixproduct(modelviewtransform[modelviewtop], modelviewmatrix[modelviewtop], modelviewtransform[modelviewtop-1]);
  finalchanged = true;
  clipspacechanged = true;
}

void MatrixManager::modelviewPop()
{
  modelviewtop --;
  
  finalchanged = true;
  clipspacechanged = true;
}

void MatrixManager::modelviewPushTranslate(float x, float y, float z)
{
  float matrix[16];
  identity(matrix);
  matrix[12]  = x;
  matrix[13]  = y;
  matrix[14] = z;

  modelviewPush(matrix);
}

void MatrixManager::modelviewPushScale(float x, float y, float z)
{
  float matrix[16];
  memset(matrix, 0, sizeof(float) * 16);
  matrix[0] = x;
  matrix[5] = y;
  matrix[10] = z;
  matrix[15] = 1.0f;
  
  modelviewPush(matrix);
}

void MatrixManager::modelviewPushRotate(float x, float y, float z, float deg)
{
  /* that is the math behind (xyz have to be normalized)
    xx xy xz
    yx yy yz
    zx zy zz
        +
    cos(alpha)
        *
    1-xx 0-xy 0-xz
    0-yx 1-yy 0-yz
    0-zx 0-zy 1-zz
        +
    sin(alpha)
        *
     0 -z  y
     z  0 -x
    -y  x  0
  */

  //I'm not sure whether this calcs much faster on vu0. I'm too lazy to try it out.

  float length = sqrt(x*x + y*y + z*z);
  x /= length;
  y /= length;
  z /= length;

  float s = sin(deg/45*atan(1)); //sin and cos accept rad instead of deg
  float c = cos(deg/45*atan(1));
  float xx = x*x;
  float xy = x*y;
  float xz = x*z;
  float yy = y*y;
  float yz = y*z;
  float zz = z*z;

  float matrix[16];
  memset(matrix, 0, sizeof(float) * 16);

  matrix[0] = xx + c* ( 1-xx );
  matrix[4] = xy + c* ( 0-xy ) + s*(-z );
  matrix[8] = xz + c* ( 0-xz ) + s*( y );
  //    [12] = 0;

  matrix[1] = xy + c* ( 0-xy ) + s*( z );
  matrix[5] = yy + c* ( 1-yy );
  matrix[9] = yz + c* ( 0-yz ) + s*(-x );
  //    [13] = 0;

  matrix[2] = xz + c* ( 0-xz ) + s*(-y );
  matrix[6] = yz + c* ( 0-yz ) + s*( x );
  matrix[10]= zz + c* ( 1-zz );
  //    [14]= 0;

  //[3]= 0;
  //[7]= 0;
  //[11]= 0;
  matrix[15]= 1;
  
  modelviewPush(matrix);
}

void MatrixManager::perspective(float * matrix)
{
  memcpy(perspectivematrix, matrix, sizeof(float) * 16);
  matrixproduct(perspectivescreenfittransform, perspectivematrix, screenfitmatrix);
  finalchanged = true;
  clipspacechanged = true;
}

void MatrixManager::perspectiveInfinite(float znear, float left, float right, float top, float bottom)
{
  memset(perspectivematrix, 0, sizeof(float) * 16);
  
  perspectivematrix[0] = 2*znear / (right - left);
  //               [1] = 0;
  //               [2] = 0;
  //               [3] = 0;
  
  //               [4] = 0;
  perspectivematrix[5] = 2*znear / (top - bottom);
  //               [6] = 0;
  //               [7] = 0;
  
  perspectivematrix[8] = (right + left) / (right - left);
  perspectivematrix[9] = (top + bottom) / (top - bottom);
  perspectivematrix[10] = -1;
  perspectivematrix[11] = -1;
  
  //               [12] = 0;
  //               [13] = 0;
  perspectivematrix[14] = -2*znear;
  //               [15] = 0;
  
  matrixproduct(perspectivescreenfittransform, perspectivematrix, screenfitmatrix);
  finalchanged = true;
  clipspacechanged = true;
}

//aspect is given in width/height
void MatrixManager::perspectiveInfinite(float znear, float aspect, float fov)
{
  float top = znear * tan((fov/45*atan(1))/2); //converts to rad and divides by 2
  perspectiveInfinite(znear, -(aspect*top) , (aspect*top), top, -top);
}

void MatrixManager::screenfit(float * matrix)
{
  memcpy(screenfitmatrix, matrix, sizeof(float) * 16);
  matrixproduct(perspectivescreenfittransform, perspectivematrix, screenfitmatrix);
  finalchanged = true;
  //the clipspace is not changed this time...
}


const float * MatrixManager::gettransform()
{
  if (finalchanged)
  {
    matrixproduct(finaltransform, modelviewtransform[modelviewtop], perspectivescreenfittransform);
    finalchanged = false;
  }

  return finaltransform;
}


const float * MatrixManager::getclipspacetransform()
{
  if (clipspacechanged)
  {
    matrixproduct(clipspacetransform, modelviewtransform[modelviewtop], perspectivematrix);
    clipspacechanged = false;
  }

  return clipspacetransform;
}


void MatrixManager::matrixproduct(float * output, const float * inputa, const float * inputb)
{
  asm __volatile__("
  lqc2 vf16, 0x00(%1)
  lqc2 vf17, 0x10(%1)
  lqc2 vf18, 0x20(%1)
  lqc2 vf19, 0x30(%1)
  lqc2 vf20, 0x00(%2)
  lqc2 vf21, 0x10(%2)
  lqc2 vf22, 0x20(%2)
  lqc2 vf23, 0x30(%2)
  
  vmulax.xyzw   ACC, vf20, vf16
  vmadday.xyzw  ACC, vf21, vf16
  vmaddaz.xyzw  ACC, vf22, vf16
  vmaddw.xyzw  vf16, vf23, vf16
  
  vmulax.xyzw   ACC, vf20, vf17
  vmadday.xyzw  ACC, vf21, vf17
  vmaddaz.xyzw  ACC, vf22, vf17
  vmaddw.xyzw  vf17, vf23, vf17
  
  vmulax.xyzw   ACC, vf20, vf18
  vmadday.xyzw  ACC, vf21, vf18
  vmaddaz.xyzw  ACC, vf22, vf18
  vmaddw.xyzw  vf18, vf23, vf18
  
  vmulax.xyzw   ACC, vf20, vf19
  vmadday.xyzw  ACC, vf21, vf19
  vmaddaz.xyzw  ACC, vf22, vf19
  vmaddw.xyzw  vf19, vf23, vf19

  sqc2 vf16, 0x00(%0)
  sqc2 vf17, 0x10(%0)
  sqc2 vf18, 0x20(%0)
  sqc2 vf19, 0x30(%0)
  "
  :
  : "r" (output), "r" (inputa), "r" (inputb)
  );

}

void MatrixManager::identity(float * matrix)
{
  std::memset(matrix, 0, 16*sizeof(float));
  matrix[0] = 1.0f;
  matrix[5] = 1.0f;
  matrix[10] = 1.0f;
  matrix[15] = 1.0f;
}


//-----------

void MatrixManager::loadmatrix()
{
  asm __volatile__("
  
  lqc2 vf16, 0x00(%0)
  lqc2 vf17, 0x10(%0)
  lqc2 vf18, 0x20(%0)
  lqc2 vf19, 0x30(%0)
  "
  :
  : "r" (gettransform())
  );
}

void MatrixManager::loadclipspacematrix()
{
  asm __volatile__("
  
  lqc2 vf16, 0x00(%0)
  lqc2 vf17, 0x10(%0)
  lqc2 vf18, 0x20(%0)
  lqc2 vf19, 0x30(%0)
  "
  :
  : "r" (getclipspacetransform())
  );
}


void MatrixManager::facetransform(void * destface, void * face)
{
  asm __volatile__("
      
  lqc2 vf20, 0x00(%1)
  lqc2 vf21, 0x10(%1)
  lqc2 vf22, 0x20(%1)

  vmulax.xyzw  ACC, vf16, vf20
  vmadday.xyzw ACC, vf17, vf20
  vmaddaz.xyzw ACC, vf18, vf20
  vmaddw.xyzw  vf20, vf19, vf20
  vdiv         Q,  vf00w, vf20w
  
  vmulax.xyzw  ACC, vf16, vf21
  vmadday.xyzw ACC, vf17, vf21
  vmaddaz.xyzw ACC, vf18, vf21
  vmaddw.xyzw  vf21, vf19, vf21
  vwaitq                       #from above
  vmulq.xyzw   vf20, vf20, Q   #from above
  vdiv         Q,  vf00w, vf21w

  vmulax.xyzw  ACC, vf16, vf22
  vmadday.xyzw ACC, vf17, vf22
  vmaddaz.xyzw ACC, vf18, vf22
  vmaddw.xyzw  vf22, vf19, vf22
  vwaitq                      #from above
  vmulq.xyzw   vf21, vf21, Q  #from above
  vdiv         Q, vf00w,  vf22w

  sqc2 vf20, 0x00(%0)
  sqc2 vf21, 0x10(%0)
  vwaitq                      #from above
  vmulq.xyzw   vf22, vf22, Q  #from above
  sqc2 vf22, 0x20(%0)
  "
  : 
  : "r"(destface), "r" (face)
  );
}

void MatrixManager::facetransformi4(void * destface, void * face)
{
  asm __volatile__("
  
  #the code looks scrabled due to optimisations
  #however this should keep both parts of the vu occupied
  
  lqc2 vf20, 0x00(%1)
  lqc2 vf21, 0x10(%1)
  lqc2 vf22, 0x20(%1)

  vmulax.xyzw  ACC, vf16, vf20
  vmadday.xyzw ACC, vf17, vf20
  vmaddaz.xyzw ACC, vf18, vf20
  vmaddw.xyzw  vf20, vf19, vf20
  vdiv         Q,  vf00w, vf20w
  
  vmulax.xyzw  ACC, vf16, vf21
  vmadday.xyzw ACC, vf17, vf21
  vmaddaz.xyzw ACC, vf18, vf21
  vmaddw.xyzw  vf21, vf19, vf21
  vwaitq                       #from above
  vmulq.xyzw   vf20, vf20, Q   #from above
  vdiv         Q,  vf00w, vf21w

  vmulax.xyzw  ACC, vf16, vf22
  vmadday.xyzw ACC, vf17, vf22
  vmaddaz.xyzw ACC, vf18, vf22
  vmaddw.xyzw  vf22, vf19, vf22
  vwaitq                      #from above
  vmulq.xyzw   vf21, vf21, Q  #from above
  vdiv         Q, vf00w,  vf22w

  vftoi4.xyzw  vf20, vf20
  vftoi4.xyzw  vf21, vf21
  vwaitq                      #from above
  vmulq.xyzw   vf22, vf22, Q  #from above
  vftoi4.xyzw  vf22, vf22


  sqc2 vf20, 0x00(%0)
  sqc2 vf21, 0x10(%0)
  sqc2 vf22, 0x20(%0)
  "
  : 
  : "r" (destface), "r" (face)
  );

}

//note: calling this 3 times is way slower that transforming a whole face at once
void MatrixManager::vertextransform(void * destvertex, void * vertex)
{
  asm __volatile__("

  lqc2 vf20, 0x00(%1)

  vmulax.xyzw  ACC, vf16, vf20
  vmadday.xyzw ACC, vf17, vf20
  vmaddaz.xyzw ACC, vf18, vf20
  vmaddw.xyzw  vf20, vf19, vf20
  vdiv         Q,  vf00w, vf20w

  vwaitq
  vmulq.xyzw   vf20, vf20, Q

  sqc2 vf20, 0x00(%0)

  "
  :
  : "r"(destvertex), "r" (vertex)
  );
}

//same as above: calling this 3 times is way slower than transforming a face
void MatrixManager::vertextransformi4(void * destvertex, void * vertex)
{
  asm __volatile__("

  lqc2 vf20, 0x00(%1)

  vmulax.xyzw  ACC, vf16, vf20
  vmadday.xyzw ACC, vf17, vf20
  vmaddaz.xyzw ACC, vf18, vf20
  vmaddw.xyzw  vf20, vf19, vf20
  vdiv         Q,  vf00w, vf20w

  vwaitq
  vmulq.xyzw   vf20, vf20, Q

  vftoi4.xyzw  vf20, vf20

  sqc2 vf20, 0x00(%0)
  "
  :
  : "r" (destvertex), "r" (vertex)
  );

}

