//    Copyright (C) 2004 Michael Mahr

//    This file is part of "lesschess".
//
//    "lesschess" is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    "lesschess" is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with "lesschess"; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "sps2ascmodel.h"

//if you are wondering where "matrixmanager" comes from... it is made globally available in defs.h, instanciated in defs.cpp, described in matrixmanager.h and implemented in matrixmanager.cpp


SPS2AscModel::SPS2AscModel(std::string _filename, float * _scale, float * _rotate, float * color, float * lightcolor, float * lightorigin)
{
  memcpy(scale, _scale, sizeof(float) * 3);
  memcpy(rotate, _rotate, sizeof(float) * 4);

  asc = new XAscModel(_filename, XAscMode(vertexnormals | facenormals));
  
  //normalize the lightorigin
  float length = sqrt(pow(lightorigin[0],2)+pow(lightorigin[1],2)+pow(lightorigin[2],2));

  AscVector origin;
  origin.x = lightorigin[0] / length;
  origin.y = lightorigin[1] / length;  
  origin.z = lightorigin[2] / length;

  //calculate the lightinfo for every vertex
  for(std::map<unsigned int, AscVector>::const_iterator h = asc->vertexnormals.begin(); h != asc->vertexnormals.end(); h++)
  {
    float dp = h->second.x * origin.x + h->second.y * origin.y + h->second.z * origin.z;
    
    float vertexshade;
    
    if (dp>0)
      vertexshade = 0;
    else
    {
      vertexshade = -dp;
    }

    //the final lightinfo consists of the origin-lightinfo
    vertexcolor[h->first].r = color[0] + lightcolor[0] * vertexshade;
    vertexcolor[h->first].g = color[1] + lightcolor[1] * vertexshade;
    vertexcolor[h->first].b = color[2] + lightcolor[2] * vertexshade;    
  }  


  //copy the data to the unbranched 
  numfaces = asc->faces.size();
  faces = new float [numfaces * 12];
  colors = new unsigned char [numfaces * 9];
  
  unsigned int k = 0;
  for (std::map<unsigned int, AscFace>::const_iterator i = asc->faces.begin(); i != asc->faces.end(); i++)
  {
  
    colors[k*9+0] = clamp(vertexcolor[i->second.a].r*255);
    colors[k*9+1] = clamp(vertexcolor[i->second.a].g*255);
    colors[k*9+2] = clamp(vertexcolor[i->second.a].b*255);
    
    colors[k*9+3] = clamp(vertexcolor[i->second.b].r*255);
    colors[k*9+4] = clamp(vertexcolor[i->second.b].g*255);
    colors[k*9+5] = clamp(vertexcolor[i->second.b].b*255);

    colors[k*9+6] = clamp(vertexcolor[i->second.c].r*255);
    colors[k*9+7] = clamp(vertexcolor[i->second.c].g*255);
    colors[k*9+8] = clamp(vertexcolor[i->second.c].b*255);    



    //std::cout << (unsigned int)(vertexcolor[i->second.a].r*255) << std::endl;
    //std::cout << (unsigned int)(colors[k*9+0]) << std::endl;
    //std::cout << "-------------" << std::endl;

    
    faces[k*12+0] = asc->verticles[i->second.a].x;
    faces[k*12+1] = asc->verticles[i->second.a].y;
    faces[k*12+2] = asc->verticles[i->second.a].z;
    faces[k*12+3] = 1.0f;
    
    faces[k*12+4] = asc->verticles[i->second.b].x;
    faces[k*12+5] = asc->verticles[i->second.b].y;
    faces[k*12+6] = asc->verticles[i->second.b].z;
    faces[k*12+7] = 1.0f;

    faces[k*12+8] = asc->verticles[i->second.c].x;
    faces[k*12+9] = asc->verticles[i->second.c].y;
    faces[k*12+10]= asc->verticles[i->second.c].z;
    faces[k*12+11]= 1.0f;
    
    k++;
  }

}

unsigned char SPS2AscModel::clamp(float input)
{
  if (input >= 255.0f)
    return 255;
  if (input <= 0.0f)
    return 0;

  return input;
}

void SPS2AscModel::draw()
{

  //allocate mem for transformation
  float destfaces[numfaces*12] __attribute__((aligned(16)));
  
  //copy over the base-data to the working data (to get it aligned);
  memcpy(destfaces, faces, numfaces * 12 * sizeof(float));
  
  //transform the model correctly
  matrixmanager.modelviewPushScale(scale[0], scale[1], scale[2]);
  matrixmanager.modelviewPushRotate(rotate[0], rotate[1], rotate[2], rotate[3]);
  
  matrixmanager.loadmatrix(); //put the current Matrix into vu0

  sps2GIFTag_t GIFTag;
  sps2GIFPackedRegister_t GIFReg;

  GIFTag.i128 = 0;
  GIFTag.s.NLOOP = numfaces * 3; //every face has 3 verticles
  GIFTag.s.PRE = 1;
  GIFTag.s.EOP = 1;
  GIFTag.s.PRIM = 11;
  GIFTag.s.FLG = GIF_FLG_PACKED;
  GIFTag.s.NREG = 0;

  SPS2_SET_GIF_REG(GIFTag, GIFTag.s.NREG++, GIF_REG_RGBAQ);
  SPS2_SET_GIF_REG(GIFTag, GIFTag.s.NREG++, GIF_REG_XYZ2);

  DPUT_EE_GIF_FIFO(GIFTag.i128);
  
 
  for (unsigned int k = 0; k<numfaces; k++)
  {
    unsigned int co = k*9;
    unsigned int fo = k*12;
  
    //transform the current face
    matrixmanager.facetransformi4(&destfaces[fo],&destfaces[fo]);
    //after this the content of destfaces no longer is "float" but "unsigned int<<4"
  
    while((*EE_GIF_STAT & (0x1F << 24))>8); //waits till fifo is _half_ empty: eliminates overflow _and_ a performance penalty

    //first vertex
    GIFReg.i128 = 0;
    GIFReg.RGBAQ.s.R = colors[co+0];
    GIFReg.RGBAQ.s.G = colors[co+1];
    GIFReg.RGBAQ.s.B = colors[co+2];
    GIFReg.RGBAQ.s.A = 0x80;
    DPUT_EE_GIF_FIFO(GIFReg.i128);
    //
    GIFReg.i128=0;
    //GIFReg.XYZ2.s.X = destfaces[fo+0];
    //GIFReg.XYZ2.s.Y = destfaces[fo+1];
    //GIFReg.XYZ2.s.Z = destfaces[fo+2];
    memcpy(&GIFReg.XYZ2.i128, &destfaces[fo+0],12); //this is a neat hack to replace the three lines above
    DPUT_EE_GIF_FIFO(GIFReg.i128);
 

    //second vertex
    GIFReg.i128 = 0;
    GIFReg.RGBAQ.s.R = colors[co+3];
    GIFReg.RGBAQ.s.G = colors[co+4];
    GIFReg.RGBAQ.s.B = colors[co+5];
    GIFReg.RGBAQ.s.A = 0x80;
    DPUT_EE_GIF_FIFO(GIFReg.i128);
    //
    GIFReg.i128=0;
    //GIFReg.XYZ2.s.X = destfaces[fo+4];
    //GIFReg.XYZ2.s.Y = destfaces[fo+5];
    //GIFReg.XYZ2.s.Z = destfaces[fo+6];
    memcpy(&GIFReg.XYZ2.i128, &destfaces[fo+4],12); //this is a neat hack to replace the three lines above
    DPUT_EE_GIF_FIFO(GIFReg.i128);

    //third vertex
    GIFReg.i128 = 0;
    GIFReg.RGBAQ.s.R = colors[co+6];
    GIFReg.RGBAQ.s.G = colors[co+7];
    GIFReg.RGBAQ.s.B = colors[co+8];
    GIFReg.RGBAQ.s.A = 0x80;
    DPUT_EE_GIF_FIFO(GIFReg.i128);
    //
    GIFReg.i128=0;
    //GIFReg.XYZ2.s.X = destfaces[fo+8];
    //GIFReg.XYZ2.s.Y = destfaces[fo+9];
    //GIFReg.XYZ2.s.Z = destfaces[fo+10];
    memcpy(&GIFReg.XYZ2.i128, &destfaces[fo+8],12); //this is a neat hack to replace the three lines above
    DPUT_EE_GIF_FIFO(GIFReg.i128);
  }

  matrixmanager.modelviewPop();
  matrixmanager.modelviewPop();


/*  the nostalgic, old, unoptimized
  for (std::map<unsigned int, AscFace>::const_iterator i = asc->faces.begin(); i != asc->faces.end(); i++)
  {
  
    //first vertex
    GIFReg.i128 = 0;
    GIFReg.RGBAQ.s.R = vertexcolor[i->second.a].r*255;
    GIFReg.RGBAQ.s.G = vertexcolor[i->second.a].g*255;
    GIFReg.RGBAQ.s.B = vertexcolor[i->second.a].b*255;
    GIFReg.RGBAQ.s.A = 0x80;
    while((*EE_GIF_STAT & (0x1F << 24))>8);
    DPUT_EE_GIF_FIFO(GIFReg.i128);
    //
    GIFReg.i128=0;
    GIFReg.XYZ2.s.X = (unsigned int)(asc->verticles[i->second.a].x * scale[0] + offset[0]) << 4;
    GIFReg.XYZ2.s.Y = (unsigned int)(asc->verticles[i->second.a].y * scale[1] + offset[1]) << 4;
    GIFReg.XYZ2.s.Z = (unsigned int)(asc->verticles[i->second.a].z * scale[2] + offset[2]);
    while((*EE_GIF_STAT & (0x1F << 24))>8);
    DPUT_EE_GIF_FIFO(GIFReg.i128);
 

    //second vertex
    GIFReg.i128 = 0;
    GIFReg.RGBAQ.s.R = vertexcolor[i->second.b].r*255;
    GIFReg.RGBAQ.s.G = vertexcolor[i->second.b].g*255;
    GIFReg.RGBAQ.s.B = vertexcolor[i->second.b].b*255;
    GIFReg.RGBAQ.s.A = 0x80;
    while((*EE_GIF_STAT & (0x1F << 24))>8);
    DPUT_EE_GIF_FIFO(GIFReg.i128);
    //
    GIFReg.i128=0;
    GIFReg.XYZ2.s.X = (unsigned int)(asc->verticles[i->second.b].x * scale[0] + offset[0]) << 4;
    GIFReg.XYZ2.s.Y = (unsigned int)(asc->verticles[i->second.b].y * scale[1] + offset[1]) << 4;
    GIFReg.XYZ2.s.Z = (unsigned int)(asc->verticles[i->second.b].z * scale[2] + offset[2]);
    while((*EE_GIF_STAT & (0x1F << 24))>8);
    DPUT_EE_GIF_FIFO(GIFReg.i128);


    //third vertex
    GIFReg.i128 = 0;
    GIFReg.RGBAQ.s.R = vertexcolor[i->second.c].r*255;
    GIFReg.RGBAQ.s.G = vertexcolor[i->second.c].g*255;
    GIFReg.RGBAQ.s.B = vertexcolor[i->second.c].b*255;
    GIFReg.RGBAQ.s.A = 0x80;
    while((*EE_GIF_STAT & (0x1F << 24))>8);
    DPUT_EE_GIF_FIFO(GIFReg.i128);
    //
    GIFReg.i128=0;
    GIFReg.XYZ2.s.X = (unsigned int)(asc->verticles[i->second.c].x * scale[0] + offset[0]) << 4;
    GIFReg.XYZ2.s.Y = (unsigned int)(asc->verticles[i->second.c].y * scale[1] + offset[1]) << 4;
    GIFReg.XYZ2.s.Z = (unsigned int)(asc->verticles[i->second.c].z * scale[2] + offset[2]);
    while((*EE_GIF_STAT & (0x1F << 24))>8);
    DPUT_EE_GIF_FIFO(GIFReg.i128);
  }
*/

}

SPS2AscModel::~SPS2AscModel()
{
  delete asc;
  
  delete faces;
  delete colors;
}
