//    Copyright (C) 2004 Michael Mahr

//    This file is part of "lesschess".
//
//    "lesschess" is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    "lesschess" is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with "lesschess"; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "sps2uvsphere.h"
#include <iostream>

SPS2UVSphere::SPS2UVSphere(unsigned int _bands, unsigned int _slices, float _radius, unsigned char _r, unsigned char _g, unsigned char _b, unsigned char _a)
{
  bands = _bands;
  slices = _slices;
  radius = _radius;
  r=_r; g=_g; b=_b; a=_a;

  //this function has been optimized, so it is kinda hard to understand 
  //  first it stores all used cos and sin values in arrays
  //  then it generates the vertex-data and puts in in class-member-arrays

  //trigonometric-caching for more speed
  const float pi = 4 * atan(1);
  float bands_sin[bands];
  float bands_cos[bands];
  float slices_sin[slices+1];
  float slices_cos[slices+1];
  
  for (unsigned int i = 1; i< (bands); i++)
  {
    bands_sin[i] = sin(i*(pi)/bands) * radius;
    bands_cos[i] = cos(i*(pi)/bands) * radius;
  }
  for (unsigned int j = 0; j< (slices); j++)
  {
    slices_sin[j] = sin(j*(2*pi)/slices);
    slices_cos[j] = cos(j*(2*pi)/slices);
  }
  slices_sin[slices] = slices_sin[0];
  slices_cos[slices] = slices_cos[0];

  //allocate memory ;  4 = xyzw
  tf_top = new float[(slices+2)*4]; //slices + 1(for top) + 1(for closing fan)
  tf_bottom = new float[(slices+2)*4]; //slices + 1(for top) + 1(for closing fan)
  ts = new float [(bands - 2) * (slices+1)*2*4]; //2 bands are created by tri-fans
  //ts is a one-dimensional array that actually is a 3-dimensional array  

  for (unsigned int k = 0; k< (bands); k++)
  {
    if (k == 0)
    {
      //draw the start-vertex of the upper triangle-fan
      tf_top[0] =  0;
      tf_top[1] =  0;
      tf_top[2] =  radius;
      tf_top[3] =  1;
     
      //all other verticles of upper triangle-fan
      for (unsigned int l = 0; l< (slices+1); l++)
      {
        //(l+1 because the start-vertex occupies the 0-position)
        tf_top[4*(l+1)+0] = bands_sin[1] * slices_cos[l];
	tf_top[4*(l+1)+1] = bands_sin[1] * slices_sin[l];
	tf_top[4*(l+1)+2] = bands_cos[1];
	tf_top[4*(l+1)+3] = 1;
      }  
    }
    else if (k == bands-1)
    {
      //first vertex of the tri-fan
      tf_bottom[0] = 0;
      tf_bottom[1] = 0;
      tf_bottom[2] = -radius;
      tf_bottom[3] = 1;
     
      //all other verticles of lower triangle-fan
      for (unsigned int l = 0; l< (slices+1); l++)
      {
        tf_bottom[4*(l+1)+0] = bands_sin[bands-1] * slices_cos[l];
	tf_bottom[4*(l+1)+1] = bands_sin[bands-1] * slices_sin[l];
	tf_bottom[4*(l+1)+2] = bands_cos[bands-1];
	tf_bottom[4*(l+1)+3] = 1;
      }  

    }
    else
    {
      //draw the body of the sphere
      for (unsigned int l = 0; l< (slices+1); l++)
      {
        ts[(slices+1)*8*(k-1)+8*l+0] = bands_sin[k] * slices_cos[l];
        ts[(slices+1)*8*(k-1)+8*l+1] = bands_sin[k] * slices_sin[l];
	ts[(slices+1)*8*(k-1)+8*l+2] = bands_cos[k];
        ts[(slices+1)*8*(k-1)+8*l+3] = 1;
	
        ts[(slices+1)*8*(k-1)+8*l+4] = bands_sin[k+1] * slices_cos[l];
        ts[(slices+1)*8*(k-1)+8*l+5] = bands_sin[k+1] * slices_sin[l];
	ts[(slices+1)*8*(k-1)+8*l+6] = bands_cos[k+1];
        ts[(slices+1)*8*(k-1)+8*l+7] = 1;
      }  
    }
  }
}

SPS2UVSphere::~SPS2UVSphere()
{

  //deallocate memory
  delete tf_top;
  delete tf_bottom;
  delete ts;
  
}


void SPS2UVSphere::draw()
{
  //the main performance-penalty in this function is the vertex-per-vertex
  // vu0-transformation. A simple enhancement would be to always transform
  // 3 vertices at once.
  //However I will now stop the project and learn how to write vu1 microcode


  const unsigned int tf_max = (slices+2) * 4; //number of tf_floats

  const unsigned int ts_max = (slices+1) * 8; //number of ts_floats
  const unsigned int bands_max = (bands - 2); //number of bands to draw
  const unsigned int bands_ts_max = bands_max * ts_max; //length of one tringle-strip in floats

  matrixmanager.loadmatrix();

  //copy the data for transformation  
  float tf_top_trans[tf_max] __attribute__((aligned(16)));
  float tf_bottom_trans[tf_max] __attribute__((aligned(16)));
  float ts_trans[bands_max * ts_max] __attribute__((aligned(16)));

  memcpy(tf_top_trans, tf_top, tf_max *sizeof(float));
  memcpy(tf_bottom_trans, tf_bottom, tf_max *sizeof(float));
  memcpy(ts_trans, ts, bands_max * ts_max *sizeof(float));

  //define all necessary tags

  sps2GIFTag_t GIFTagCol;  //gif-tag for color
  GIFTagCol.i128 = 0;
  GIFTagCol.s.NLOOP = 1; //only set color once
  GIFTagCol.s.PRE = 1;
  GIFTagCol.s.EOP= 1;
  GIFTagCol.s.PRIM = 13; //primitive-type is irrelevant
  GIFTagCol.s.FLG = GIF_FLG_PACKED;
  GIFTagCol.s.NREG = 0;
  SPS2_SET_GIF_REG(GIFTagCol, GIFTagCol.s.NREG++, GIF_REG_RGBAQ);

  sps2GIFTag_t GIFTagTf;  //gif-tag for triangle-fans
  GIFTagTf.i128 = 0;
  GIFTagTf.s.NLOOP = (slices+2);
  GIFTagTf.s.PRE = 1;
  GIFTagTf.s.EOP= 1;
  GIFTagTf.s.PRIM = 13; //triangle fan
  GIFTagTf.s.FLG = GIF_FLG_PACKED;
  GIFTagTf.s.NREG = 0;
  SPS2_SET_GIF_REG(GIFTagTf, GIFTagTf.s.NREG++, GIF_REG_XYZ2);
  
  sps2GIFTag_t GIFTagTs;  //gif-tag for triangle-strips
  GIFTagTs.i128 = 0;
  GIFTagTs.s.NLOOP = (slices+1)*2;
  GIFTagTs.s.PRE = 1;
  GIFTagTs.s.EOP= 1;
  GIFTagTs.s.PRIM = 12; //triangle strip
  GIFTagTs.s.FLG = GIF_FLG_PACKED;
  GIFTagTs.s.NREG = 0;
  SPS2_SET_GIF_REG(GIFTagTs, GIFTagTs.s.NREG++, GIF_REG_XYZ2);
  
  sps2GIFPackedRegister_t GIFReg;
  
  //----
  
  //set color

  DPUT_EE_GIF_FIFO(GIFTagCol.i128);
  
  GIFReg.i128 = 0;
  GIFReg.RGBAQ.s.R = r;
  GIFReg.RGBAQ.s.G = g;
  GIFReg.RGBAQ.s.B = b;
  GIFReg.RGBAQ.s.A = a;

  DPUT_EE_GIF_FIFO(GIFReg.i128);
  
  GIFReg.i128 = 0;

  

  //draw upper triangle-fan
  DPUT_EE_GIF_FIFO(GIFTagTf.i128);
  for (unsigned int i = 0; i < tf_max; i+=4)
  {
    matrixmanager.vertextransformi4(&tf_top_trans[i], &tf_top_trans[i]);
    
    memcpy(&GIFReg.XYZ2.i128, &tf_top_trans[i], 12);

    DPUT_EE_GIF_FIFO(GIFReg.i128);
  }
  

  //draw lower triangle-fan
  DPUT_EE_GIF_FIFO(GIFTagTf.i128);
  for (unsigned int i = 0; i < tf_max; i+=4)
  {
    matrixmanager.vertextransformi4(&tf_bottom_trans[i], &tf_bottom_trans[i]);    
    
    memcpy(&GIFReg.XYZ2.i128, &tf_bottom_trans[i], 12);

    DPUT_EE_GIF_FIFO(GIFReg.i128);
  }
  

  //draw the triangle-strips
  for (unsigned int i = 0; i < bands_ts_max; i+=ts_max)
  {
    DPUT_EE_GIF_FIFO(GIFTagTs.i128);
    
    for (unsigned int j = 0; j < ts_max; j+=4)
    {
      unsigned int ts_all_max = i+j;
      
      matrixmanager.vertextransformi4(&ts_trans[ts_all_max], &ts_trans[ts_all_max]);
    
      memcpy(&GIFReg.XYZ2.i128, &ts_trans[ts_all_max], 12);

      DPUT_EE_GIF_FIFO(GIFReg.i128);
    }
    
  }

  
}

